"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client
import kaskada
import kaskada.api.v1alpha.staged_file_pb2 as staged_pb

import datetime
import grpc
import hashlib
import io
import os
import random
import requests
import pandas as pd
import tempfile


def list_staged_files(
    search: str = None, client: Client = None
) -> staged_pb.ListStagedFilesResponse:
    """
    Lists staged files

    Args:
        search (str, optional): Name to search staged files. Defaults to None.
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        staged_pb.ListStagedFilesResponse: List of Staged Files
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        kaskada.validate_client(client)
        req = staged_pb.ListStagedFilesRequest(search=search)
        return client.stagingStub.ListStagedFiles(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def get_staged_file(
    staged_file, client: Client = None
) -> staged_pb.GetStagedFileResponse:
    """
    Gets a staged file given a file ID.

    Args:
        file (Union[staged_pb.CreateStagedFileResponse, staged_pb.StagedFile,  str]): Staged file object
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        staged_pb.GetStagedFileResponse: Staged File
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    file_id = None
    if isinstance(staged_file, staged_pb.CreateStagedFileResponse):
        file_id = staged_file.file.file_id
    elif isinstance(staged_file, staged_pb.StagedFile):
        file_id = staged_file.file_id
    elif isinstance(staged_file, str):
        file_id = staged_file
    else:
        raise Exception(
            "invalid parameters passed. please provide the staged file ID or create staged file response"
        )

    try:
        kaskada.validate_client(client)
        req = staged_pb.GetStagedFileRequest(file_id=file_id)
        return client.stagingStub.GetStagedFile(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def upload_staged_data(
    staged_file: staged_pb.CreateStagedFileResponse,
    f: io.BufferedReader,
    client: Client = None,
):
    """
    Uploads a generic buffered reader to a staged file

    Args:
        staged_file (staged_pb.CreateStagedFileResponse): The result of the create_staged_file call.
        f (io.BufferedReader): The buffered reader to stage.
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        None (if successful)
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    headers = {"X-Amz-Meta-Filename": staged_file.file.file_name}
    http_response = requests.put(staged_file.url, headers=headers, data=f)
    if http_response.status_code != 200:
        raise Exception(
            "Unable to upload file. Resulted in status code: {} and message: {}".format(
                http_response.status_code, http_response.content
            )
        )


def create_staged_file_api(
    file_name: str, content_id: str, client: Client = None
) -> staged_pb.CreateStagedFileResponse:
    """
    Creates a staged file given the file name and content id (hash)

    Args:
        file_name (str): The name of the file (used for metadata)
        content_id (str, optional): The content id of the file. Defaults to None.
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        staged_pb.CreateStagedFileResponse: Staged File
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT
    try:
        kaskada.validate_client(client)
        req = staged_pb.CreateStagedFileRequest(
            file=staged_pb.StagedFile(file_name=file_name, content_id=content_id)
        )
        return client.stagingStub.CreateStagedFile(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def create_staged_file(
    source, client: Client = None
) -> staged_pb.CreateStagedFileResponse:
    """
    Entry point to create a staged file

    Args:
        source ([str, pandas.Dataframe]): The source to create the staged file from
        client (Kaskada.Client, optional): The Kaskada Client. Defaults to global client.

    Raises:
        Exception: Invalid parameters (only path or dataframe) is accepted

    Returns:
        staged_pb.CreateStagedFileResponse: Staged File
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    if isinstance(source, str):
        f = open(source, "rb")
        file_name = os.path.basename(f.name)
        # Generate the content ID which is the MD5 hash of the file
        content_id = hashlib.md5(f.read()).hexdigest()
        staged_file = create_staged_file_api(file_name, content_id, client=client)
        f.close()
        f = open(source, "rb")
        upload_staged_data(staged_file, f, client=client)
        return staged_file
    elif isinstance(source, pd.DataFrame):
        # Content ID is the Data Frame MD5 hash which is computed off the values
        content_id = hashlib.md5(pd.util.hash_pandas_object(source).values).hexdigest()
        # Generate a name for the dataframe as the current time + random number.parquet
        current_time = datetime.datetime.now().strftime("%Y-%m-%d")
        base_name = "{}_{}.parquet".format(current_time, random.randint(0, 1000))
        staged_file = create_staged_file_api(base_name, content_id, client=client)
        with tempfile.TemporaryDirectory() as tmpdirname:
            file_name = os.path.join(tmpdirname, staged_file.file.file_name)
            # write to tempfile in `tmpdirname`
            source.to_parquet(
                path=file_name,
                engine="pyarrow",
                allow_truncated_timestamps=True,
                use_deprecated_int96_timestamps=True,
            )
            # upload tempfile from `tmpdirname`
            upload_staged_data(staged_file, open(file_name, "rb"))
            return staged_file
    else:
        raise Exception(
            "invalid parameters passed. please provide the path or dataframe as the first parameter"
        )



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""
from typing import Union

from kaskada.client import Client
import kaskada
import kaskada.api.v1alpha.view_pb2 as view_pb

import grpc


def get_view_name(
    view: Union[view_pb.View, view_pb.CreateViewResponse, view_pb.GetViewResponse, str]
) -> str:
    view_name = None
    if isinstance(view, view_pb.View):
        view_name = view.view_name
    elif isinstance(view, view_pb.CreateViewResponse) or isinstance(
        view, view_pb.GetViewResponse
    ):
        view_name = view.view.view_name
    elif isinstance(view, str):
        view_name = view

    if view_name is None:
        raise Exception(
            "invalid view parameter provided. \
                the view parameter must be the view object, view response from the SDK, or the view name"
        )

    return view_name


def list_views(search: str = None, client: Client = None) -> view_pb.ListViewsResponse:
    """
    Lists all the views the user has access to

    Args:
        search (str, optional): The search parameter to filter list by. Defaults to None.
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        view_pb.ListViewsResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        kaskada.validate_client(client)
        req = view_pb.ListViewsRequest(
            search=search,
        )
        return client.viewStub.ListViews(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def get_view(
    view: Union[view_pb.View, view_pb.CreateViewResponse, view_pb.GetViewResponse, str],
    client: Client = None,
) -> view_pb.GetViewResponse:
    """
    Gets a view by name

    Args:
        view (Union[view_pb.View, view_pb.CreateViewResponse, view_pb.GetViewResponse, str]): The target view object
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        view_pb.GetViewResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        view_name = get_view_name(view)
        kaskada.validate_client(client)
        req = view_pb.GetViewRequest(view_name=view_name)
        return client.viewStub.GetView(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def create_view(
    view_name: str, expression: str, client: Client = None
) -> view_pb.CreateViewResponse:
    """
    Creates a view with a name and expression

    Args:
        view_name (str): The view name
        expression (str): The view fenl expression
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        view_pb.CreateViewResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        kaskada.validate_client(client)
        req = view_pb.CreateViewRequest(
            view=view_pb.View(view_name=view_name, expression=expression)
        )
        return client.viewStub.CreateView(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def delete_view(
    view: Union[view_pb.View, view_pb.CreateViewResponse, view_pb.GetViewResponse, str],
    client: Client = None,
    force: bool = False,
) -> view_pb.DeleteViewResponse:
    """
    Deletes a view

    Args:
        view (Union[view_pb.View, view_pb.CreateViewResponse, view_pb.GetViewResponse, str]): The target view object
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        view_pb.DeleteViewResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        view_name = get_view_name(view)
        kaskada.validate_client(client)
        req = view_pb.DeleteViewRequest(view_name=view_name, force=force)
        return client.viewStub.DeleteView(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)



"""
Copyright (C) 2022 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""
import logging
from domonic.ext.html5lib_ import getTreeBuilder
from domonic.html import a, button, div, pre, script, style, table, td, th, tr
from domonic.utils import Utils
import html5lib
import inspect
import io
import pkg_resources
import pprint
import sys
import kaskada.formatters_shared as shared
from kaskada.formatters_helpers import (
    appendChildIfNotNone,
    appendHtmlObjTableRowIfAttrExists,
    get_classname,
    get_properties,
    get_schema_dataframe,
    html_table_row,
    html_obj_table_row,
    html_obj_id_row,
)

_LOGGER = logging.getLogger(__name__)


def in_ipython() -> bool:
    """
    Checks to see if running in an iPython environment
    """
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


def get_query_response_content(resp_obj):
    resultsExist = False
    details = table(_class="kda_table")

    state_enum_descriptor = resp_obj.DESCRIPTOR.enum_types_by_name.get("State")
    state = state_enum_descriptor.values_by_number.get(resp_obj.state).name
    details.appendChild(html_table_row("state", state.replace("STATE_", "")))

    if hasattr(resp_obj, "metrics") and resp_obj.state > 1:
        metrics = shared.get_metrics_html(resp_obj.metrics)
        details.appendChild(html_table_row("metrics", metrics))

    if resp_obj.HasField("parquet"):
        nested_table = table(_class="kda_table")
        details.appendChild(html_table_row("parquet", nested_table))
        nested_table.appendChild(tr(th("index"), th("path")))
        for i in range(len(resp_obj.parquet.paths)):
            path = resp_obj.parquet.paths[i]
            appendChildIfNotNone(nested_table, tr(td(pre(i)), td(a(path, _href=path))))
        resultsExist = True
    elif resp_obj.HasField("redis_bulk"):
        nested_table = table(_class="kda_table")
        details.appendChild(html_table_row("redis_bulk", nested_table))
        nested_table.appendChild(tr(th("index"), th("path")))
        for i in range(len(resp_obj.redis_bulk.paths)):
            path = resp_obj.redis_bulk.paths[i]
            appendChildIfNotNone(nested_table, tr(td(pre(i)), td(a(path, _href=path))))
        resultsExist = True

    (
        can_execute,
        has_errors,
        nested_details,
        diagnostics,
    ) = shared.get_analysis_and_diagnostics_tables(resp_obj)
    details.appendChild(html_table_row("analysis", nested_details))

    schema = None
    if resp_obj.HasField("analysis"):
        schema_df = get_schema_dataframe(resp_obj.analysis)
        if schema_df is not None:
            schema = convert_df_to_domonic(schema_df)
            details.appendChild(html_obj_table_row("schema", "(see Schema tab)"))

    if resultsExist and hasattr(resp_obj, "config"):
        appendHtmlObjTableRowIfAttrExists(details, resp_obj.config, "data_token_id")
        if resp_obj.config.HasField("slice_request"):
            details.appendChild(
                html_table_row(
                    "slice_request",
                    shared.get_slice_request_html(resp_obj.config.slice_request),
                )
            )
        else:
            details.appendChild(
                html_table_row("slice_request", shared.get_slice_request_html(None))
            )

    appendChildIfNotNone(
        details, shared.get_request_details_table_row_if_exists(resp_obj)
    )

    return has_errors, {
        "Details": details,
        "Schema": schema,
        "Diagnostics": diagnostics,
    }


# handles Create and Get responses for Table, View, Materialization
# also handles LoadDataResponse
def generic_response_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    details = table(_class="kda_table")
    schema_df = None
    if hasattr(obj, "table") and obj.HasField("table"):
        table_details, schema_df = shared.get_table_html_and_schema_df(obj.table)
        details.appendChild(html_table_row("table", table_details))
    elif hasattr(obj, "view") and obj.HasField("view"):
        view_details, schema_df = shared.get_view_html_and_schema_df(obj.view)
        details.appendChild(html_table_row("view", view_details))
    elif hasattr(obj, "materialization") and obj.HasField("materialization"):
        mat_details, schema_df = shared.get_materialization_html_and_schema_df(
            obj.materialization
        )
        details.appendChild(html_table_row("materialization", mat_details))
    else:
        appendHtmlObjTableRowIfAttrExists(details, obj, "data_token_id")

    # pull of analysis
    diagnostics = None
    has_errors = False
    if hasattr(obj, "analysis"):
        (
            can_execute,
            has_errors,
            nested_details,
            diagnostics,
        ) = shared.get_analysis_and_diagnostics_tables(obj.analysis)
        details.appendChild(html_table_row("analysis", nested_details))

    appendChildIfNotNone(details, shared.get_request_details_table_row_if_exists(obj))

    schema = convert_df_to_domonic(schema_df) if schema_df is not None else None
    title_to_set_active = "Diagnostics" if has_errors else "Details"
    return str(
        tab_panel(
            [
                ("Details", details),
                ("Schema", schema),
                ("Diagnostics", diagnostics),
                ("Raw", shared.get_raw_html(obj)),
            ],
            title_to_set_active=title_to_set_active,
        )
    )


# handles Table, View, Materialization, and other generic objects
def generic_object_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    if get_classname(type(obj)) == "kaskada.api.v1alpha.table_pb2.Table":
        details, schema_df = shared.get_table_html_and_schema_df(obj)
    elif get_classname(type(obj)) == "kaskada.api.v1alpha.view_pb2.View":
        details, schema_df = shared.get_view_html_and_schema_df(obj)
    elif (
        get_classname(type(obj))
        == "kaskada.api.v1alpha.materialization_pb2.Materialization"
    ):
        details, schema_df = shared.get_materialization_html_and_schema_df(obj)
    else:
        details, schema_df = shared.get_generic_object_html_and_schema_df(obj)

    schema = convert_df_to_domonic(schema_df) if schema_df is not None else None
    return str(
        tab_panel(
            [
                ("Details", details),
                ("Schema", schema),
                ("Raw", shared.get_raw_html(obj)),
            ]
        )
    )


# handles the Schema object
def schema_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    schema_df = get_schema_dataframe(obj)
    schema = convert_df_to_domonic(schema_df) if schema_df is not None else None

    return str(tab_panel([("Schema", schema), ("Raw", shared.get_raw_html(obj))]))


# handles the DataType object
def data_type_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    result_type, schema_df = shared.get_result_type_and_schema_df(obj)
    title_to_set_active = "Schema" if schema_df is not None else ""
    schema = convert_df_to_domonic(schema_df) if schema_df is not None else None

    return str(
        tab_panel(
            [
                ("Result Type", pre(result_type)),
                ("Schema", schema),
                ("Raw", shared.get_raw_html(obj)),
            ],
            title_to_set_active,
        )
    )


# handles Delete responses for Table, View, Materialization
def response_delete_html_formatter(del_resp):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    request_details = shared.get_request_details_table_row_if_exists(del_resp)

    del_resp_type = type(del_resp).__name__
    del_obj = del_resp_type.replace("Response", "")

    details = table(_class="kda_table")
    details.appendChild(html_obj_table_row(del_obj, "success"))
    details.appendChild(request_details)

    return str(
        tab_panel([("Details", details), ("Raw", pre(pprint.pformat(del_resp)))])
    )


# handles List responses for Tables, Views, Materializations
def response_list_html_formatter(list_resp):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    list_resp_type = type(list_resp).__name__
    list_key = list_resp_type.replace("Response", "").replace("List", "").lower()

    details = table(_class="kda_table")

    # after pulling the request_id, only the item list should remain
    props = get_properties(list_resp)
    props.pop("request_details")
    if len(props) == 1:
        list_value = getattr(list_resp, list_key)
        list_props = props[list_key]
        list_type = type(list_value)
        nested_table = table(_class="kda_table")
        details.appendChild(html_table_row(list_key, nested_table))
        nested_table.appendChild(tr(th("index"), th("name")))

        if (
            get_classname(list_type)
            == "google.protobuf.pyext._message.RepeatedCompositeContainer"
        ):
            for i in range(len(list_value)):
                appendChildIfNotNone(
                    nested_table, html_obj_id_row(str(i), list_value[i], list_props[i])
                )
    else:  # when list is empty
        details.appendChild(html_table_row(list_key, Utils.escape("<empty>")))

    appendChildIfNotNone(
        details, shared.get_request_details_table_row_if_exists(list_resp)
    )

    return str(
        tab_panel([("Details", details), ("Raw", shared.get_raw_html(list_resp))])
    )


def proto_list_html_formatter(list_resp):
    item_count = len(list_resp)
    if item_count == 0:
        return str(shared.get_raw_html(list_resp))

    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    details = table(_class="kda_table")

    # peek at the type of items in the list
    item_class = get_classname(type(list_resp[0]))

    if item_class == "kaskada.errdetails.v1alpha.fenl_diagnostics_pb2.FenlDiagnostic":
        for i in range(item_count):
            fenl_diag = list_resp[i]
            details.appendChild(html_obj_table_row(i, fenl_diag.formatted))
    else:
        details.appendChild(tr(th("index"), th("name")))
        for i in range(item_count):
            appendChildIfNotNone(
                details, html_obj_id_row(i, list_resp[i], get_properties(list_resp[i]))
            )

    return str(
        tab_panel([("Details", details), ("Raw", shared.get_raw_html(list_resp))])
    )


def slice_request_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    return str(
        tab_panel(
            [
                ("Details", shared.get_slice_request_html(obj)),
                ("Raw", shared.get_raw_html(obj)),
            ]
        )
    )


def query_metrics_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()
    return str(
        tab_panel(
            [
                ("Details", shared.get_metrics_html(obj)),
                ("Raw", shared.get_raw_html(obj)),
            ]
        )
    )


def query_response_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    has_errors, content = get_query_response_content(obj)
    title_to_set_active = "Diagnostics" if has_errors else "Details"
    return str(
        tab_panel(
            [
                ("Details", content["Details"]),
                ("Schema", content["Schema"]),
                ("Diagnostics", content["Diagnostics"]),
                ("Raw", shared.get_raw_html(obj)),
            ],
            title_to_set_active=title_to_set_active,
        )
    )


def fenlmagic_query_result_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    has_errors, content = get_query_response_content(obj.query_response)
    details_table = content["Details"]
    details_table.appendChild(html_obj_table_row("query", obj.query))

    raw = {
        "query": obj.query,
        "query_response": obj.query_response,
    }

    dataframe = None
    if obj.dataframe is not None:
        buffer = io.StringIO()
        obj.dataframe.info(buf=buffer, verbose=False)
        raw["dataframe"] = buffer.getvalue().replace(
            "<class 'pandas.core.frame.DataFrame'>\n", ""
        )
        dataframe = convert_df_to_domonic(obj.dataframe, 10, True)

    title_to_set_active = "Details"
    if has_errors:
        title_to_set_active = "Diagnostics"
    elif dataframe is not None:
        title_to_set_active = "Dataframe"

    return str(
        tab_panel(
            [
                ("Dataframe", dataframe),
                ("Details", content["Details"]),
                ("Schema", content["Schema"]),
                ("Diagnostics", content["Diagnostics"]),
                ("Raw", shared.get_raw_html(raw)),
            ],
            title_to_set_active=title_to_set_active,
        )
    )


def entity_filter_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    details = table(_class="kda_table")
    filter_type = type(obj).__name__
    if filter_type == "EntityPercentFilter":
        details.appendChild(html_obj_table_row("percent", obj.percent))

    return str(tab_panel([("Details", details), ("Raw", shared.get_raw_html(obj))]))


def analysis_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    (
        can_execute,
        has_errors,
        details,
        diagnostics,
    ) = shared.get_analysis_and_diagnostics_tables(obj)
    title_to_set_active = "Diagnostics" if has_errors else "Details"
    return str(
        tab_panel(
            [
                ("Details", details),
                ("Diagnostics", diagnostics),
                ("Raw", shared.get_raw_html(obj)),
            ],
            title_to_set_active=title_to_set_active,
        )
    )


def fenl_diagnostics_html_formatter(obj):
    output_custom_css_and_javascript_if_output_wrapped_in_iframe()

    diagnostics, has_errors = shared.get_fenl_diagnostics_html_and_has_errors(obj)
    details = table(html_table_row("fenl_diagnostics", diagnostics))
    return str(tab_panel([("Details", details), ("Raw", shared.get_raw_html(obj))]))


def convert_df_to_domonic(df, max_rows=None, show_dimensions=False):
    parser = html5lib.HTMLParser(tree=getTreeBuilder())
    return parser.parse(
        df.to_html(max_rows=max_rows, notebook=True, show_dimensions=show_dimensions)
    )


def tab_panel(content_list, title_to_set_active=""):
    """
    Creates a tab-panel for the passed content list.

    Args:
        content_list (list of tuple of (str, Any)): A list of tuples, where the the first value
        in the tuple is the tab title, and the second is the tab contents.

        title_to_set_active str: The name of the tab title to set active when the panel loads. If
        unset, the first tab will be set active.

    Returns:
        domonic.html.div: A div html element wrapping the tabs and content.  Pass this value to
        str() to get the string html output.
    """
    if len(content_list) == 0:
        return None

    tab_links = div(_class="kda_tab")
    root_div = div(tab_links)

    for titleContent in content_list:
        title = titleContent[0]
        content = titleContent[1]
        if content is None:
            continue
        if title_to_set_active == "":
            title_to_set_active = title
        lower_title = title.lower()

        tab_button_class = (
            "kda_tablinks kda_active"
            if title_to_set_active == title
            else "kda_tablinks"
        )
        tab_links.appendChild(
            button(
                title,
                _class=tab_button_class,
                _onclick=f"openTab(event, '{lower_title}')",
            )
        )

        scroll_div = div(_class="kda_scroll_container").html(content)
        content_div = div(_class=f"{lower_title} kda_tabcontent").html(scroll_div)
        content_div.style.display = "block" if title_to_set_active == title else "none"
        root_div.appendChild(content_div)

    return root_div


def output_custom_css_and_javascript():
    """
    Outputs our custom css and javascript to the current cell
    """
    import IPython

    css = pkg_resources.resource_string(__name__, "formatters.css").decode("utf-8")
    js = pkg_resources.resource_string(__name__, "formatters.js").decode("utf-8")
    IPython.core.display.display(IPython.core.display.HTML(str(style(css))))
    IPython.core.display.display(
        IPython.core.display.HTML(str(script(js, _type="text/javascript")))
    )


def cell_output_is_wrapped_in_iframe():
    """
    Based on the environment that the code is running inside of, the cell output may be
    automatically wrapped in an iFrame.
    For example, in google colab: "the output of each cell is hosted in a separate iframe
    sandbox with limited access to the global notebook environment"
    In enviroments like these, we need to include our custom css & javascript in every output.

    Returns:
        bool: True if the environment will automatically wrap the cell output in an iFrame, otherwise False.
    """
    from IPython import get_ipython
    if "google.colab" in str(get_ipython()):
        return True
    else:
        return False


def output_custom_css_and_javascript_if_output_wrapped_in_iframe():
    """
    Outputs our custom css and javascript to the cell output if the output for the cell
    will be wrapped in an iFrame.  (i.e. in google colab)
    """
    if cell_output_is_wrapped_in_iframe():
        output_custom_css_and_javascript()


def try_init():
    try:
        from IPython import get_ipython
        # the following command will throw an exception in non-iPython environments
        html_formatter = get_ipython().display_formatter.formatters["text/html"]

        # dynamically assign formatters to kaskada protobuf types
        mods = sys.modules.copy()
        for key in mods:
            if key.endswith("_grpc"):
                continue
            if key.startswith("kaskada."):
                for cls in inspect.getmembers(mods[key], inspect.isclass):
                    classname = get_classname(cls[1])
                    if (
                        classname
                        == "kaskada.api.v1alpha.compute_pb2.StreamQueryResponse"
                    ):
                        html_formatter.for_type(
                            classname, query_response_html_formatter
                        )  # kda_table formatter for QueryResponse
                    elif classname == "kaskada.fenl.v1alpha.schema_pb2.Schema":
                        html_formatter.for_type(classname, schema_html_formatter)
                    elif classname == "kaskada.fenl.v1alpha.schema_pb2.DataType":
                        html_formatter.for_type(classname, data_type_html_formatter)
                    elif classname == "kaskada.prepare.v1alpha.slice_pb2.SliceRequest":
                        html_formatter.for_type(classname, slice_request_html_formatter)
                    elif (
                        classname
                        == "kaskada.errdetails.v1alpha.fenl_diagnostics_pb2.FenlDiagnostics"
                    ):
                        html_formatter.for_type(
                            classname, fenl_diagnostics_html_formatter
                        )
                    elif classname == "kaskada.api.v1alpha.shared_pb2.Analysis":
                        html_formatter.for_type(classname, analysis_html_formatter)
                    elif "Delete" in classname and "Response" in classname:
                        html_formatter.for_type(
                            classname, response_delete_html_formatter
                        )  # generic formatter for Delete responses
                    elif "List" in classname and "Response" in classname:
                        html_formatter.for_type(
                            classname, response_list_html_formatter
                        )  # generic formatter for List responses
                    elif "Response" in classname:
                        html_formatter.for_type(
                            classname, generic_response_html_formatter
                        )  # generic formatter for other responses
                    else:
                        html_formatter.for_type(
                            classname, generic_object_html_formatter
                        )  # generic formatter for all other objects

        # the following types don't normally exist when the library first loads
        html_formatter.for_type(
            "kaskada.api.v1alpha.compute_pb2.Metrics", query_metrics_html_formatter
        )
        html_formatter.for_type(
            "kaskada.materialization.MaterializationView", generic_object_html_formatter
        )
        html_formatter.for_type(
            "kaskada.materialization.RedisAIDestination", generic_object_html_formatter
        )
        html_formatter.for_type(
            "kaskada.compute.EntityPercentFilter", entity_filter_html_formatter
        )
        html_formatter.for_type(
            "fenlmagic.QueryResult", fenlmagic_query_result_html_formatter
        )

        # additional non-kaskada types we want to assign formatters to
        html_formatter.for_type(
            "google.protobuf.pyext._message.RepeatedCompositeContainer",
            proto_list_html_formatter,
        )

        if not cell_output_is_wrapped_in_iframe():
            # load our custom css into the page
            output_custom_css_and_javascript()

    except Exception as e:
        _LOGGER.debug("Init exception: {}".format(e))



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""
from typing import List, Union

import grpc
from kaskada.api.v1alpha.compute_pb2 import WithView
from kaskada.slice_filters import EntityPercentFilter, SliceFilter
from kaskada.client import Client
import kaskada
import kaskada.api.v1alpha.materialization_pb2 as material_pb
import kaskada.prepare.v1alpha.slice_pb2 as slice_pb


class RedisAIDestination(object):
    db: int = 0
    host: str = None
    port: int = 0

    def __init__(self, db: int, host: str, port: int):
        """
        Kaskada Redis AI Destination object

        Args:
            db (int): The RedisAI database number to write to. Between 0 and 15 inclusive.
            host (str): The RedisAI instance hostname
            port (int): The RedisAI instance port
        """
        self.db = db
        self.host = host
        self.port = port


class MaterializationView(object):
    name: str = None
    expression: str = None

    def __init__(self, name: str, expression: str):
        """
        Kaskada Materialization View

        Args:
            name (str): The name of the view
            expression (str): The fenl expression to compute
        """
        self.name = name
        self.expression = expression


def create_materialization(
    name: str,
    query: str,
    destination: RedisAIDestination,
    views: List[MaterializationView] = [],
    slice_filter: SliceFilter = None,
    client: Client = None,
) -> material_pb.CreateMaterializationResponse:
    """
    Creates a materialization

    Args:
        name (str):
            The Materialization's Name.
        query (str):
            A Fenl expression to compute.
        destination (RedisAIDestination):
            RedisAI instance
        views (list, optional):
            List of MaterializationView. Defaults to [].
        slice_filter (SliceFilter, optional):
            Enables slice filter. Currently, only slice entity percent filters are supported. Defaults to None.

    Returns:
        material_pb.CreateMaterializationResponse: [description]
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        slice_request = None
        if slice_filter is not None:
            if isinstance(slice_filter, EntityPercentFilter):
                slice_request = slice_pb.SliceRequest(
                    percent=slice_pb.SliceRequest.PercentSlice(
                        percent=slice_filter.get_percent()
                    )
                )
            else:
                raise Exception(
                    "invalid slice filter provided. only EntityPercentFilter is supported"
                )

        req = material_pb.CreateMaterializationRequest(
            materialization=material_pb.Materialization(
                materialization_name=name,
                query=query,
                with_views=to_with_views(views),
                redis_a_i=material_pb.Materialization.RedisAI(
                    db=destination.db, host=destination.host, port=destination.port
                ),
                slice=slice_request,
            )
        )

        return client.materializationStub.CreateMaterialization(
            req, metadata=client.get_metadata()
        )
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def delete_materialization(
    materialization: Union[
        material_pb.Materialization,
        material_pb.CreateMaterializationResponse,
        material_pb.GetMaterializationResponse,
        str,
    ],
    client: Client = None,
) -> material_pb.DeleteMaterializationResponse:
    """
    Deletes a materialization by name

    Args:
        materialization (
            Union[
                material_pb.Materialization,
                material_pb.CreateMaterializationResponse,
                material_pb.GetMaterializationResponse,
                str]):
                    The target materialization object
        client (Client, optional):
            The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        material_pb.DeleteMaterializationResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        materialization_name = get_materialization_name(materialization)
        kaskada.validate_client(client)
        req = material_pb.DeleteMaterializationRequest(
            materialization_name=materialization_name
        )
        return client.materializationStub.DeleteMaterialization(
            req, metadata=client.get_metadata()
        )
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def get_materialization(
    materialization: Union[
        material_pb.Materialization,
        material_pb.CreateMaterializationResponse,
        material_pb.GetMaterializationResponse,
        str,
    ],
    client: Client = None,
) -> material_pb.GetMaterializationResponse:
    """
    Gets a materialization by name

    Args:
        materialization (
            Union[
                material_pb.Materialization,
                material_pb.CreateMaterializationResponse,
                material_pb.GetMaterializationResponse,
                str]):
                    The target materialization object
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        material_pb.GetMaterializationResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        materialization_name = get_materialization_name(materialization)
        kaskada.validate_client(client)
        req = material_pb.GetMaterializationRequest(
            materialization_name=materialization_name
        )
        return client.materializationStub.GetMaterialization(
            req, metadata=client.get_metadata()
        )
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def list_materializations(
    search: str = None, client: Client = None
) -> material_pb.ListMaterializationsResponse:
    """
    Lists all the materializations the user has access to

    Args:
        search (str, optional): The search parameter to filter list by. Defaults to None.
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        material_pb.ListMaterializationsResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        kaskada.validate_client(client)
        req = material_pb.ListMaterializationsRequest(search=search)
        return client.materializationStub.ListMaterializations(
            req, metadata=client.get_metadata()
        )
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def get_materialization_name(materialization) -> str:
    name = None
    if isinstance(materialization, str):
        name = materialization
    elif isinstance(
        materialization, material_pb.CreateMaterializationResponse
    ) or isinstance(materialization, material_pb.GetMaterializationResponse):
        name = materialization.materialization.materialization_name
    elif isinstance(materialization, material_pb.Materialization):
        name = materialization.materialization_name
    else:
        raise Exception(
            "invalid materialization parameter provided. the materialization parameter must be the materialization \
                object, response from SDK or the materialization name"
        )
    return name


def to_with_views(views: List[MaterializationView]) -> List[WithView]:
    with_views = []
    for v in views:
        with_views.append(WithView(name=v.name, expression=v.expression))
    return with_views



"""
Copyright (C) 2022 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from domonic.html import pre, table
from domonic.utils import Utils
import pprint

from kaskada.formatters_helpers import (
    appendChildren,
    appendHtmlObjTableRowIfAttrExists,
    convert_fenl_datatype,
    get_datetime,
    get_properties,
    get_schema_dataframe,
    html_table_row,
    html_obj_table_row,
)

'''
NOTES:

Use `hasattr(obj, prop) to test if an object has a property`
Use `obj.HasField(prop) to test if an object has a specific one-of`
'''


def appendTimes(parent, obj):
    if hasattr(obj, "create_time"):
        create_time = get_datetime(getattr(obj, "create_time")).isoformat()
        parent.appendChild(html_obj_table_row("create_time", create_time))
    if hasattr(obj, "update_time"):
        update_time = get_datetime(getattr(obj, "update_time")).isoformat()
        parent.appendChild(html_obj_table_row("update_time", update_time))


def get_analysis_and_diagnostics_tables(obj):
    can_execute = (
        obj.analysis.can_execute if hasattr(obj, "analysis") else obj.can_execute
    )
    details = table(
        html_obj_table_row("can_execute", str(can_execute)), _class="kda_table"
    )
    diagnostics = None
    has_errors = False
    if (
        hasattr(obj, "fenl_diagnostics")
        and len(obj.fenl_diagnostics.fenl_diagnostics) > 0
    ):
        details.appendChild(
            html_obj_table_row("fenl_diagnostics", "(see Diagnostics tab)")
        )
        diagnostics, has_errors = get_fenl_diagnostics_html_and_has_errors(
            obj.fenl_diagnostics
        )
    return can_execute, has_errors, details, diagnostics


def get_fenl_diagnostics_html_and_has_errors(obj):
    diagnostics = table(_class="kda_table")
    for i in range(len(obj.fenl_diagnostics)):
        fenl_diag = obj.fenl_diagnostics[i]
        diagnostics.appendChild(html_obj_table_row(i, fenl_diag.formatted))
    return diagnostics, obj.num_errors > 0


def get_generic_object_html_and_schema_df(obj):
    props = get_properties(obj)
    obj_type = type(obj).__name__.lower()

    # try to pop off obj_id and obj_name first
    obj_name_key = f"{obj_type}_name"
    obj_name = props.pop(obj_name_key, None)

    # also remove create & update time from props
    props.pop("create_time", None)
    props.pop("update_time", None)

    rows = [
        # first row is item name
        html_obj_table_row(obj_name_key, obj_name),
        # stop showing item ids since they aren't actionable by the user
        # html_obj_table_row(obj_id_key, obj_id),
    ]

    # add remainder of fields
    for key in props:
        rows.append(html_obj_table_row(key, props[key]))

    schema_df = get_schema_dataframe(obj)
    details = table(_class="kda_table")
    appendChildren(details, rows)
    if schema_df is not None:
        details.appendChild(html_obj_table_row("schema", "(see Schema tab)"))
    appendTimes(details, obj)
    return details, schema_df


def get_materialization_html_and_schema_df(obj):
    schema_df = get_schema_dataframe(obj)
    details = table(_class="kda_table")
    appendHtmlObjTableRowIfAttrExists(details, obj, "materialization_name")
    appendHtmlObjTableRowIfAttrExists(details, obj, "query")

    if hasattr(obj, "redis_a_i") and obj.HasField("redis_a_i"):
        redis_a_i = table(_class="kda_table")
        appendHtmlObjTableRowIfAttrExists(redis_a_i, obj.redis_a_i, "host")
        appendHtmlObjTableRowIfAttrExists(redis_a_i, obj.redis_a_i, "port")
        appendHtmlObjTableRowIfAttrExists(redis_a_i, obj.redis_a_i, "db")
        details.appendChild(html_table_row("redis_a_i", redis_a_i))

    if hasattr(obj, "slice"):
        details.appendChild(html_table_row("slice", get_slice_request_html(obj.slice)))

    if schema_df is not None:
        details.appendChild(html_obj_table_row("schema", "(see Schema tab)"))

    appendTimes(details, obj)
    return details, schema_df


def get_metrics_html(obj):
    metrics = table(_class="kda_table")

    time_preparing = (
        obj.time_preparing.ToMilliseconds() if hasattr(obj, "time_preparing") else 0
    )
    metrics.appendChild(html_table_row("time_preparing", f"{time_preparing / 1000}s"))

    time_computing = (
        obj.time_computing.ToMilliseconds() if hasattr(obj, "time_computing") else 0
    )
    metrics.appendChild(html_table_row("time_computing", f"{time_computing / 1000}s"))

    metrics.appendChild(html_table_row("output_files", obj.output_files))
    return metrics


def get_raw_html(obj):
    return pre(pprint.pformat(obj).replace("\\n", ""))


def get_request_details_table_row_if_exists(obj):
    if hasattr(obj, "request_details") and hasattr(obj.request_details, "request_id"):
        nested_table = table(
            html_obj_table_row("request_id", obj.request_details.request_id),
            _class="kda_table",
        )
        return html_table_row("request_details", nested_table)
    return None


def get_result_type_and_schema_df(obj):
    result_type = Utils.escape(convert_fenl_datatype(obj))
    schema_df = None
    if obj.HasField("struct"):
        schema_df = get_schema_dataframe(obj.struct)
        if schema_df is not None:
            result_type += " (see Schema tab)"
    return result_type, schema_df


def get_slice_request_html(obj):
    details = table(_class="kda_table")
    if obj is not None and obj.HasField("percent"):
        details.appendChild(html_obj_table_row("percent", obj.percent.percent))
    else:
        details.appendChild(html_obj_table_row("None", "(full dataset used for query)"))
    return details


def get_table_html_and_schema_df(obj):
    schema_df = get_schema_dataframe(obj)
    details = table(_class="kda_table")
    appendHtmlObjTableRowIfAttrExists(details, obj, "table_name")
    appendHtmlObjTableRowIfAttrExists(details, obj, "entity_key_column_name")
    appendHtmlObjTableRowIfAttrExists(details, obj, "time_column_name")
    if obj.HasField("subsort_column_name"):
        details.appendChild(
            html_obj_table_row("subsort_column_name", obj.subsort_column_name.value)
        )

    if hasattr(obj, "table_source") and obj.HasField("table_source"):
        if hasattr(obj.table_source, "iceberg") and obj.table_source.HasField("iceberg"):
            iceberg = table(_class="kda_table")
            appendHtmlObjTableRowIfAttrExists(iceberg, obj.table_source.iceberg, "table_name")
            appendHtmlObjTableRowIfAttrExists(iceberg, obj.table_source.iceberg, "namespace")
            if hasattr(obj.table_source.iceberg, "config") and obj.table_source.iceberg.HasField("config"):
                config = table(_class="kda_table")
                if hasattr(obj.table_source.iceberg.config, "glue") and obj.table_source.iceberg.config.HasField(
                    "glue"
                ):
                    glue = table(_class="kda_table")
                    if hasattr(
                        obj.table_source.iceberg.config.glue, "warehouse"
                    ) and obj.table_source.iceberg.config.glue.HasField("warehouse"):
                        warehouse = table(_class="kda_table")
                        appendHtmlObjTableRowIfAttrExists(
                            warehouse, obj.table_source.iceberg.config.glue.warehouse, "bucket"
                        )
                        appendHtmlObjTableRowIfAttrExists(
                            warehouse, obj.table_source.iceberg.config.glue.warehouse, "region"
                        )
                        appendHtmlObjTableRowIfAttrExists(
                            warehouse, obj.table_source.iceberg.config.glue.warehouse, "prefix"
                        )
                        glue.appendChild(html_table_row("warehouse", warehouse))
                    config.appendChild(html_table_row("glue", glue))
                iceberg.appendChild(html_table_row("config", config))
            details.appendChild(html_table_row("iceberg", iceberg))

    appendHtmlObjTableRowIfAttrExists(details, obj, "version")
    if schema_df is not None:
        details.appendChild(html_obj_table_row("schema", "(see Schema tab)"))
    appendTimes(details, obj)
    return details, schema_df


def get_view_html_and_schema_df(obj):
    details = table(_class="kda_table")
    appendHtmlObjTableRowIfAttrExists(details, obj, "view_name")
    appendHtmlObjTableRowIfAttrExists(details, obj, "expression")
    schema_df = None
    if hasattr(obj, "result_type"):
        result_type, schema_df = get_result_type_and_schema_df(obj.result_type)
        if result_type is not None and result_type != "":
            details.appendChild(html_obj_table_row("result_type", result_type))

    appendTimes(details, obj)
    return details, schema_df



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client
from typing import List
import kaskada
import kaskada.api.v1alpha.compute_pb2 as pb
import kaskada.formatters
from kaskada.slice_filters import SliceFilter
from google.protobuf.timestamp_pb2 import Timestamp
from typing import Union
import datetime

import grpc


def query(
    query: str,
    with_views: List[pb.WithView] = [],
    result_behavior: str = "all-results",
    response_as: str = "parquet",
    data_token_id: str = None,
    dry_run: bool = False,
    changed_since_time: Union[str, datetime.datetime, None] = None,
    limits: pb.QueryRequest.Limits = None,
    slice_filter: SliceFilter = None,
    experimental: bool = False,
    client: Client = None,
) -> pb.StreamQueryResponse:
    """
    Performs a query

    Args:
        query (str): The query to perform
        with_views (List[pb.WithView], optional):
            A list of views to use in the query, in addition to the views stored in the system.
        result_behavior (str, optional):
            Determines which results are returned. Either "all-results" (default), or "final-results" which returns
            only the final values for each entity.
        responsed_as (str, optional):
            Determines how the response is returned.  Either "parquet" (default) or "redis-bulk".
            Note: if "redis-bulk", result_behavior is assumed to be "final-results".
        data_token_id (str, optional):
            Enables repeatable queries. Queries performed against the same dataToken are always ran the same input data.
        dry_run(bool, optional):
            When `True`, the query is validated and if there are no errors, the resultant schema is returned.
            No actual computation of results is performed.
        changed_since_time (datetime.datetime, optional):
            Configure the inclusive datetime after which results will be output.
        limits (pb.QueryRequest.Limits, optional):
            Configure limits on the output set.
        slice_filter (SliceFilter, optional):
            Enables slice filter. Currently, only slice entity percent filters are supported. Defaults to None.
        experimental(bool, optional):
            When `True`, then experimental features are allowed. Data returned when using this flag is not
            guaranteed to be correct.
        client (Client, optional):
            The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        compute_pb.QueryResponse: Response from the API
    """

    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    if slice_filter is None:
        slice_filter = kaskada.KASKADA_DEFAULT_SLICE

    timestamp = None
    if isinstance(changed_since_time, str):
        timestamp = Timestamp()
        timestamp.FromJsonString(changed_since_time)
    elif isinstance(changed_since_time, datetime.datetime):
        timestamp = Timestamp()
        timestamp.FromDatetime(changed_since_time)
    elif changed_since_time is not None:
        raise Exception(
            "Invalid type for `changed_since_time`. Expected `str` or `datetime.datetime`."
        )

    try:
        kaskada.validate_client(client)

        request_args = {
            "data_token_id": data_token_id,
            "dry_run": dry_run,
            "experimental_features": experimental,
            "changed_since_time": timestamp,
            "limits": limits,
            "query": query,
            "with_views": with_views,
        }

        if result_behavior == "final-results":
            request_args["result_behavior"] = "RESULT_BEHAVIOR_FINAL_RESULTS"
        else:
            request_args["result_behavior"] = "RESULT_BEHAVIOR_ALL_RESULTS"
        if response_as == "redis-bulk":
            request_args["redis_bulk"] = {}
        else:
            request_args["parquet"] = {}

        if slice_filter is not None:
            request_args["slice_request"] = slice_filter.to_request()

        in_ipython = kaskada.formatters.in_ipython()
        if in_ipython:
            from IPython.display import display, clear_output

            request_args["stream_metrics"] = True

        request = pb.StreamQueryRequest(**request_args)
        response = pb.StreamQueryResponse()

        responses = client.computeStub.StreamQuery(
            request, metadata=client.get_metadata()
        )
        for resp in responses:
            response.MergeFrom(resp)
            if in_ipython:
                clear_output(wait=True)
                display(response)

        if in_ipython:
            clear_output(wait=True)
        return response
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)



"""
Copyright (C) 2022 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from datetime import datetime
from domonic.html import b, pre, td, tr
import pandas
import re

import kaskada.fenl.v1alpha as kaskada_fenl_pb2

'''
NOTES:

Use `hasattr(obj, prop) to test if an object has a property`
Use `obj.HasField(prop) to test if an object has a specific one-of`
'''

kaskada_fenl_primitive_types = {
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_BOOL: "bool",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_DURATION_MICROSECOND: "duration_us",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_DURATION_MILLISECOND: "duration_ms",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_DURATION_NANOSECOND: "duration_ns",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_DURATION_SECOND: "duration_s",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_F16: "f16",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_F32: "f32",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_F64: "f64",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_I16: "i16",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_I32: "i32",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_I64: "i64",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_I8: "i8",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_INTERVAL_DAY_TIME: "interval_days",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_INTERVAL_YEAR_MONTH: "interval_months",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_NULL: "null",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_STRING: "string",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_TIMESTAMP_MICROSECOND: "timestamp_us",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_TIMESTAMP_MILLISECOND: "timestamp_ms",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_TIMESTAMP_NANOSECOND: "timestamp_ns",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_TIMESTAMP_SECOND: "timestamp_s",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_U16: "u16",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_U32: "u32",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_U64: "u64",
    kaskada_fenl_pb2.schema_pb2.DataType.PRIMITIVE_TYPE_U8: "u8",
}


def get_item(obj, keys, default=""):
    """
    Safely recursively gets an item from a dict, otherwise returns the default value

    Args:
        obj (dict of {str : Any}): Dictionary of (nested) key item pairs
        keys (list of str): The list of keys use to navigate to the desired item in the dictionary.
        default (str, optional): The default value to return if the item isn't found. Defaults to "".

    Returns:
        string: Either the found value or the default.
    """
    while len(keys) > 0:
        key = keys.pop(0)
        if key in obj:
            obj = obj[key]
        else:
            return default
    return obj


def appendChildIfNotNone(parent, child):
    if child is not None:
        parent.appendChild(child)


def appendChildren(parent, children):
    for child in children:
        appendChildIfNotNone(parent, child)


def html_obj_table_row(key, value):
    """
    Returns a html table row, where the key is bolded, and the value is output as code.
    """
    if value is None or value == "":
        return None
    return tr(td(b(key)), td(pre(value)))


def html_table_row(key, value):
    """
    Returns a html table row, where the key is bolded, and the value is output as is.
    """
    return tr(td(b(key)), td(value))


def html_obj_id_row(index, obj, props):
    obj_type = type(obj).__name__.lower()
    obj_name_key = f"{obj_type}_name"
    obj_name = get_item(props, [obj_name_key])
    return tr(td(pre(index)), td(pre(obj_name)))


def appendHtmlObjTableRowIfAttrExists(parent, obj, attr):
    if hasattr(obj, attr):
        parent.appendChild(html_obj_table_row(attr, getattr(obj, attr)))


def appendHtmlObjTableRowIfHasField(parent, obj, field):
    if obj.HasField(field):
        parent.appendChild(html_obj_table_row(field, getattr(obj, field)))


def convert_fenl_schema_field(field):
    """
    Converts a fenl field to a string representation.

    Args:
        field (kaskada_fenl_pb2.schema_pb2.Schema_Field) : the input field

    Returns:
        str: The string representation of the fenl Field
    """
    return field.name, convert_fenl_datatype(field.data_type)


def convert_fenl_datatype(data_type):
    """
    Converts a fenl dataType to a string representation.

    Args:
        data_type (kaskada_fenl_pb2.schema_pb2.DataType) : the input dataType

    Returns:
        str: The string representation of the fenl DataType
    """
    type_string = None
    if data_type.HasField("primitive"):
        type_string = kaskada_fenl_primitive_types.get(data_type.primitive, "<unknown>")
    elif data_type.HasField("struct"):
        type_string = "<struct>"
    elif data_type.HasField("window"):
        type_string = "<window>"
    return type_string


def get_schema_dataframe(obj):
    """
    Gets a dataframe from a schema object if it exists

    Args:
        obj (kaskada.fenl.v1alpha.schema_pb2.Schema | Any):
            Either a schema object or an object that has a schema or analysis object at the root level

    Returns:
        pandas.dataframe | None:
            Either a dataframe describing the schemea or None if no schema exists in the passed object.
    """
    if hasattr(obj, "schema"):
        obj = obj.schema

    if (not hasattr(obj, "fields")) or len(obj.fields) == 0:
        return None
    fields = {}
    for idx, field in enumerate(obj.fields):
        name, typeString = convert_fenl_schema_field(field)
        fields[idx] = [name, typeString]
    return pandas.DataFrame.from_dict(
        data=fields, orient="index", columns=["column_name", "column_type"]
    )


def get_datetime(pb2_timestamp):
    """
    Converts a google.protobuf.timestamp_pb2 into a python datetime object.

    Args:
        pb2_timestamp (google.protobuf.timestamp_pb2): The protobuf timestamp.

    Returns:
        datetime: The python datetime.
    """
    return datetime.fromtimestamp(pb2_timestamp.seconds + pb2_timestamp.nanos / 1e9)


def get_classname(obj):
    return f"{obj.__module__}.{obj.__name__}"


def get_properties(obj):
    """
    Converts a protobuf response into a simple python dictionary.

    Args:
        obj (Any): A deserialized protobuf object.

    Returns:
        dict of {str : Any}: A nested set of keys and values with structure matching the protobuf input.
    """
    results = {}
    for propName in dir(obj):
        if re.search(r"^[a-z]", propName) is None:
            continue
        propVal = getattr(obj, propName)
        propType = type(propVal)
        prop_class = get_classname(propType)
        if propType.__name__ in ("method"):
            continue
        elif propType in (int, float, bool):
            results[propName] = propVal
        elif propType is str:
            if propVal != "":
                results[propName] = propVal
        elif prop_class == "google.protobuf.pyext._message.RepeatedCompositeContainer":
            if len(propVal) > 0:
                results[propName] = [get_properties(subObj) for subObj in propVal]
        elif prop_class == "google.protobuf.timestamp_pb2.Timestamp":
            results[propName] = get_datetime(propVal).isoformat()
        elif prop_class.startswith("pandas."):
            results[propName] = propVal
        else:
            results[propName] = get_properties(propVal)
    return results



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

import unicodedata
import re


def normalize_col_name(column: str) -> str:
    """
    Normalize column name by replacing invalid characters with underscore
    strips accents and make lowercase
    """
    n = re.sub(r"[ ,;{}()\n\t=]+", "_", column.lower())
    return unicodedata.normalize("NFKD", n).encode("ASCII", "ignore").decode()



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""
from typing import Union

from kaskada.client import Client
import google.protobuf.wrappers_pb2 as wrappers
import kaskada
import kaskada.staging as ks
import kaskada.api.v1alpha.shared_pb2 as shared_pb
import kaskada.api.v1alpha.staged_file_pb2 as staged_pb
import kaskada.api.v1alpha.table_pb2 as table_pb

import grpc
import pandas as pd


class ExternalTable(object):
    pass


class IcebergTable(ExternalTable):
    def __init__(
        self, catalog_bucket: str, catalog_prefix: str, namespace: str, table_name: str
    ):
        self.catalog_bucket = catalog_bucket
        self.catalog_prefix = catalog_prefix
        self.namespace = namespace
        self.table_name = table_name
        # Default region to us-west-2
        self.region = "us-west-2"


def get_table_name(
    table: Union[
        table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str
    ]
) -> str:
    """
    Gets the table name from either the table protobuf, create table response, get table response, or a string

    Args:
        table (Union[table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str]):
            The target table object

    Returns:
        str: The table name (None if unable to match)
    """
    table_name = None
    if isinstance(table, table_pb.Table):
        table_name = table.table_name
    elif isinstance(table, table_pb.CreateTableResponse) or isinstance(
        table, table_pb.GetTableResponse
    ):
        table_name = table.table.table_name
    elif isinstance(table, str):
        table_name = table

    if table_name is None:
        raise Exception(
            "invalid table parameter provided. the table parameter must be the table object, table response from the \
                SDK, or the table name"
        )

    return table_name


def list_tables(
    search: str = None, client: Client = None
) -> table_pb.ListTablesResponse:
    """
    Lists all tables the user has access to

    Args:
        search (str, optional): The search parameter to filter list by. Defaults to None.
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        table_pb.ListTablesResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        kaskada.validate_client(client)
        req = table_pb.ListTablesRequest(
            search=search,
        )
        return client.tableStub.ListTables(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def get_table(
    table: Union[
        table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str
    ],
    client: Client = None,
) -> table_pb.GetTableResponse:
    """
    Gets a table by name

    Args:
        table (Union[Table, CreateTableResponse, GetTableResponse, str]): The target table object
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        table_pb.GetTableResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        table_name = get_table_name(table)
        kaskada.validate_client(client)
        req = table_pb.GetTableRequest(table_name=table_name)
        return client.tableStub.GetTable(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def create_table(
    table_name: str,
    time_column_name: str,
    entity_key_column_name: str,
    subsort_column_name: str = None,
    grouping_id: str = None,
    source: ExternalTable = None,
    client: Client = None,
) -> table_pb.CreateTableResponse:
    """
    Creates a table

    Args:
        table_name (str):
            The name of the table
        time_column_name (str):
            The time column
        entity_key_column_name (str):
            The entity key column
        subsort_column_name (str, optional):
            The subsort column. Defaults to None and Kaskada will generate a subsort column for the data.
        grouping_id (str, optional):
            The grouping id. Defaults to None.
        client (Client, optional):
            The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        table_pb.CreateTableResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        kaskada.validate_client(client)
        table_args = {
            "table_name": table_name,
            "time_column_name": time_column_name,
            "entity_key_column_name": entity_key_column_name,
            "grouping_id": grouping_id,
        }
        if subsort_column_name:
            table_args["subsort_column_name"] = wrappers.StringValue(
                value=subsort_column_name
            )

        if source:
            if isinstance(source, IcebergTable):
                table_args["table_source"] = {
                    "iceberg": {
                        "config": {
                            "glue": {
                                "warehouse": {
                                    "bucket": source.catalog_bucket,
                                    "region": source.region,
                                    "prefix": source.catalog_prefix,
                                }
                            }
                        },
                        "namespace": source.namespace,
                        "table_name": source.table_name,
                    }
                }
            else:
                raise Exception("invalid external source provided")
        req = table_pb.CreateTableRequest(table=table_pb.Table(**table_args))
        return client.tableStub.CreateTable(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def delete_table(
    table: Union[
        table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str
    ],
    client: Client = None,
    force: bool = False,
) -> table_pb.DeleteTableResponse:
    """
    Deletes a table referenced by name

    Args:
        table (Union[Table, CreateTableResponse, GetTableResponse, str]): The target table object
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        table_pb.DeleteTableResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    try:
        table_name = get_table_name(table)
        kaskada.validate_client(client)
        req = table_pb.DeleteTableRequest(table_name=table_name, force=force)
        return client.tableStub.DeleteTable(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def load_data(
    table: Union[
        table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str
    ],
    staged_file: Union[
        staged_pb.StagedFile,
        staged_pb.CreateStagedFileResponse,
        staged_pb.GetStagedFileResponse,
        str,
    ],
    client: Client = None,
) -> table_pb.LoadDataResponse:
    """
    Loads data to a table from a staged file

    Args:
        table (Union[Table, CreateTableResponse, GetTableResponse, str]): The target table object
        staged_file (Union[StagedFile, CreateStagedFileResponse, GetStagedFileResponse, Str]): The staged file object
        client (Client, optional): The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Raises:
        Exception: No valid table parameter provided
        Exception: No valid staged file ID parameter provided

    Returns:
        table_pb.LoadDataResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    staged_file_id = None
    if isinstance(staged_file, staged_pb.StagedFile):
        staged_file_id = staged_file.file_id
    elif isinstance(staged_file, staged_pb.CreateStagedFileResponse) or isinstance(
        staged_file, staged_pb.GetStagedFileResponse
    ):
        staged_file_id = staged_file.file.file_id
    elif isinstance(staged_file, str):
        staged_file_id = staged_file

    if staged_file_id is None:
        raise Exception(
            "invalid staged file parameter provided. the second parameter must be the staged file object, \
                staged file response from the SDK, or the staged file ID"
        )

    try:
        table_name = get_table_name(table)
        kaskada.validate_client(client)
        req = table_pb.LoadDataRequest(
            table_name=table_name,
            file_id=staged_file_id,
        )
        return client.tableStub.LoadData(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def upload_file(
    table: Union[
        table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str
    ],
    file_path: str,
    client: Client = None,
):
    """
    Uploads a file directly to a table. Reads the file and uses the staging area to load the data.

    Args:
        table (Union[table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str]):
            The target table object
        file_path (str):
            The source local file path
        client (Client, optional):
            The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        LoadDataResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT
    staged_file = ks.create_staged_file(file_path, client=client)
    return load_data(table, staged_file)


def upload_from_s3(
    table: Union[
        table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str
    ],
    access_key: str = None,
    secret: str = None,
    virtual_hosted: str = None,
    region: str = None,
    bucket: str = None,
    key: str = None,
    client: Client = None,
):
    """
    Uploads a file from an external S3 source to a table. Credentials are required and are not persisted in any way.

    Args:
        table (Union[table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str]):
            The target table object
        access_key (str, optional):
            Amazon Web Services Access Key. Used only for authentication and is not persisted.
        secret (str, optional):
            Amazon Web Services Secret Key. Used only for authentication and is not persisted.
        virtual_hosted (str, optional):
            The Amazon Virtual Hosted URI to the file.
            See: https://docs.aws.amazon.com/AmazonS3/latest/userguide/VirtualHosting.html.
            If not provided, the region + bucket + key will be used.
        region (str, optional):
            The target region of the S3 bucket. Only used if virtual_hosted is not provided.
        bucket (str, optional):
            The target S3 bucket. Only used if virtual_hosted is not provided.
        key (str, optional):
            The target key to the object in the S3 bucket. Only used if virtual_hosted is not provided.
        client (Client, optional):
            The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        LoadDataResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT

    external_s3_file = {}
    if virtual_hosted is not None:
        external_s3_file["virtual_hosted_style"] = virtual_hosted
    else:
        external_s3_file["bucket_region_key"] = {
            "bucket": bucket,
            "region": region,
            "key": key,
        }
    try:
        table_name = get_table_name(table)
        kaskada.validate_client(client)

        if access_key and secret:
            credentials = shared_pb.AWSCredential(
                access_key={
                    "aws_access_key_id": access_key,
                    "aws_secret_access_key": secret,
                }
            )
            external_s3_file["aws_credential"] = credentials
        elif access_key and not secret:
            raise ValueError("Access Key provided but Secret was not")
        elif not access_key and secret:
            raise ValueError("Secret provided but Access Key was not")
        external_s3_file = shared_pb.ExternalS3File(**external_s3_file)
        req = table_pb.LoadDataRequest(
            table_name=table_name, external_s3_file=external_s3_file
        )
        return client.tableStub.LoadData(req, metadata=client.get_metadata())
    except grpc.RpcError as e:
        kaskada.handleGrpcError(e)
    except Exception as e:
        kaskada.handleException(e)


def upload_dataframe(
    table: Union[
        table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str
    ],
    df: pd.DataFrame,
    client: Client = None,
):
    """
    Uploads a dataframe directly to a table. Writes the dataframe to a local file and loads the data to the table

    Args:
        table (Union[table_pb.Table, table_pb.CreateTableResponse, table_pb.GetTableResponse, str]):
            The target table object
        df (pd.DataFrame):
            The source Pandas dataframe
        client (Client, optional):
            The Kaskada Client. Defaults to kaskada.KASKADA_DEFAULT_CLIENT.

    Returns:
        LoadDataResponse: Response from the API
    """
    if client is None:
        client = kaskada.KASKADA_DEFAULT_CLIENT
    staged_file = ks.create_staged_file(df, client=client)
    return load_data(table, staged_file)



from typing import Dict, List, Union


class SliceFilter(object):
    def to_request(self):
        raise NotImplementedError()


class EntityPercentFilter(SliceFilter):
    # The lower limit allowed for percents of entities to filter through (out of a 100)
    # The percent must be greater than or equal to this limit
    LOWER_LIMIT = 0.1
    # The upper limit allowed for percents of entities to filter through (out of a 100).
    # The percent must be less than or equal to this limit
    UPPER_LIMIT = 100.0

    def __init__(self, percent: float):
        """
        Initializes an entity percent filter. Only one filter is allowed per query request.

        Args:
            percent (float): The percent of entities to filter.
                             Must be GTE to LOWER_LIMIT and LTE to UPPER_LIMIT.
        """
        if (
            percent < EntityPercentFilter.LOWER_LIMIT
            or percent > EntityPercentFilter.UPPER_LIMIT
        ):
            raise Exception(
                "percent must be between {} and {}".format(
                    EntityPercentFilter.LOWER_LIMIT, EntityPercentFilter.UPPER_LIMIT
                )
            )
        super().__init__()
        self.percent = percent

    def to_request(self) -> Dict[str, Dict[str, float]]:
        return {"percent": {"percent": self.get_percent()}}

    def get_percent(self) -> float:
        """
        Returns the percent of entities to filter

        Returns:
            float: The percent between (0, 100)
        """
        return self.percent


class EntityFilter(SliceFilter):
    entities: List[str] = []

    def __init__(self, entities: Union[List[int], List[str]]):
        """
        Initializes a filter for a list of entity keys. Only one filter is allowed per query request.
        Although the filter accepts any format, each entity key will be encoded as a string using
        a string formatter.

        Example:
        - Entity Key: "user_id1" encodes to "user_id1"
        - Entity Key: 123 encodes to "123"

        Args:
            entity (str): The entity keys to filter values by.
        """
        self.entities = []
        for e in entities:
            self.entities.append("{}".format(e))

    def to_request(self):
        """Generates the dictionary version of the request

        Returns:
            _type_: _description_
        """
        return {
            "entity_keys": {
                "entity_keys": self.entities,
            }
        }



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from kaskada.client import Client
from kaskada.slice_filters import SliceFilter
import kaskada.formatters

import grpc
import IPython
import os
import sys
import traceback

import kaskada.errdetails.v1alpha.fenl_diagnostics_pb2 as kaskada_error_details_pb2

from grpc_status import rpc_status
from google.rpc import error_details_pb2

# Module global client configured at initialization time. By default, all API calls utilize the global client config
KASKADA_DEFAULT_CLIENT = None
# Client ID environment variable to authenticate with the API
KASKADA_CLIENT_ID_ENV = "KASKADA_CLIENT_ID"
# Client Secret environment variable to authenticate with the API
KASKADA_CLIENT_SECRET_ENV = "KASKADA_CLIENT_SECRET"
# Public API endpoint for Kaskada service
KASKADA_API_ENDPOINT = "api.kaskada.com:50051"
KASKADA_API_ENDPOINT_ENV = "KASKADA_API_ENDPOINT"
# Authentication endpoint for Kaskada service. JWTs exchange authentication endpoint.
KASKADA_AUTH_ENDPOINT = "prod-kaskada.us.auth0.com"
KASKADA_AUTH_ENDPOINT_ENV = "KASKADA_AUTH_ENDPOINT"
# Authentication audience for JWTs creation.
KASKADA_AUTH_AUDIENCE = "https://api.prod.kaskada.com"
KASKADA_AUTH_AUDIENCE_ENV = "KASKADA_AUTH_AUDIENCE"
# Use SSL connection to connect to the API
KASKADA_IS_SECURE_ENV = "KASKADA_IS_SECURE"
# Demo Kaskada Client ID has read-only permissions
KASKADA_DEMO_CLIENT_ID_ENV = "KASKADA_DEMO_CLIENT_ID"
# Trial Kaskada Client ID has standard permissions
KASKADA_TRIAL_CLIENT_ID_ENV = "KASKADA_TRIAL_CLIENT_ID"

# The slicing parameter passed to query by default. By default, this is None.
KASKADA_DEFAULT_SLICE = None


def showtraceback(self, running_compiled_code=True):
    traceback_lines = traceback.format_exception(*sys.exc_info())
    message = traceback_lines[-1]
    sys.stderr.write(message)


IPython.core.interactiveshell.InteractiveShell.showtraceback = showtraceback


def init(
    client_id: str = None,
    client_secret: str = None,
    exchange_endpoint: str = None,
    audience: str = None,
    endpoint: str = None,
    is_secure: bool = None,
    jwt: str = None,
):
    """
    Initializes the Kaskada library with a client. Creating a client will update the default module default client.

    Authentication:
    1. Client ID and Client Secret
    2. JWT
    3. Demo Client ID or Client Secret provided

    Args:
        client_id (str, optional): Kaskada Client ID.
        client_secret (str, optional): Kaskada Client Secret.
        exchange_endpoint (str, optional): Authentication endpoint to exchange JWTs.
        audience (str, optional): Authentication audience.
        endpoint (str, optional): API Endpoint.
        is_secure (bool, optional): Use SSL connection. Defaults to True.
        jwt (str, optional):  Jwt for preauthorized clients.
    Returns:
        Client: Kaskada Client
    """
    kaskada.formatters.try_init()

    # Use the provided Client ID and Client Secret to login
    if client_id is not None:
        assert client_secret is not None, "invalid client secret"
    # No Client ID or JWT was provided. Only Demo and Trial modes supported.
    elif jwt is None:
        demo_client_id = os.getenv(KASKADA_DEMO_CLIENT_ID_ENV)
        if demo_client_id is not None:
            client_id = demo_client_id

        trial_client_id = os.getenv(KASKADA_TRIAL_CLIENT_ID_ENV)
        if trial_client_id is not None:
            client_id = trial_client_id

        if client_id is None:
            client_id = os.getenv(KASKADA_CLIENT_ID_ENV)
            if client_id is None:
                raise ValueError(
                    "Unable to read Kaskada Client ID. \n" +
                    "Please provide client_id key word argument or set env var KASKADA_CLIENT_ID."
                )

            if client_secret is None:
                client_secret = os.getenv(KASKADA_CLIENT_SECRET_ENV)
            if client_secret is None:
                raise ValueError(
                    "Unable to read Kaskada Client Secret. \n" +
                    "Please provide client_secret key word argument or set env var KASKADA_CLIENT_SECRET."
                )

    if exchange_endpoint is None:
        exchange_endpoint = os.getenv(KASKADA_AUTH_ENDPOINT_ENV, KASKADA_AUTH_ENDPOINT)

    if audience is None:
        audience = os.getenv(KASKADA_AUTH_AUDIENCE_ENV, KASKADA_AUTH_AUDIENCE)

    if endpoint is None:
        endpoint = os.getenv(KASKADA_API_ENDPOINT_ENV, KASKADA_API_ENDPOINT)

    if is_secure is None:
        is_secure = os.getenv(KASKADA_IS_SECURE_ENV, "true") != "false"

    global KASKADA_DEFAULT_CLIENT
    KASKADA_DEFAULT_CLIENT = Client(
        client_id=client_id,
        client_secret=client_secret,
        jwt=jwt,
        exchange_endpoint=exchange_endpoint,
        audience=audience,
        endpoint=endpoint,
        is_secure=is_secure
    )


def set_default_slice(slice: SliceFilter):
    """
    Sets the default slice used in every query

    Args:
        slice (SliceFilter): SliceFilter to set the default
    """
    global KASKADA_DEFAULT_SLICE
    KASKADA_DEFAULT_SLICE = slice


def handleGrpcError(rpc_error: grpc.Call):
    """
    All methods calling `handleGrpcError` do so after detecting a RpcError.  Currently
    all our grpc calls perform a `UnaryUnaryMultiCallable` method, docs linked here:
    https://grpc.github.io/grpc/python/_modules/grpc.html#UnaryUnaryMultiCallable

    Calling a `UnaryUnaryMultiCallable` method results in one of the following:
        Returns:
            The response value for the RPC.

        Raises:
            RpcError: Indicating that the RPC terminated with non-OK status. The
            raised RpcError will also be a grpc.Call for the RPC affording the RPC's
            metadata, status code, and details.

    Methods on a `grpc.Call` object include the following.  Full docs here:
    https://grpc.github.io/grpc/python/_modules/grpc.html#Call

        code(): The StatusCode value for the RPC.
        details(): The details string of the RPC.

    The `grpc_status.rpc_status.from_call(call)` method helps extract details
    objects from `grpc.Call` objects. Docs here:
    https://grpc.github.io/grpc/python/_modules/grpc_status/rpc_status.html#from_call

        Args:
            call: A grpc.Call instance.

        Returns:
            A google.rpc.status.Status message representing the status of the RPC.

        Raises:
            ValueError: If the gRPC call's code or details are inconsistent with the
            status code and message inside of the google.rpc.status.Status.
    """
    errorMessage = "An error occurred in your request.\n\tError Code: {}\n".format(
        rpc_error.code().name
    )
    try:
        status = rpc_status.from_call(rpc_error)
        if status:
            unpacked = []
            errorMessage += "\tError Message: {}\n".format(status.message)
            for detail in status.details:
                try:
                    val = unpack_details(detail)
                    unpacked.append(val)
                except Exception:
                    # maybe report back to wren in the future
                    pass
            if len(unpacked) > 0:
                errorMessage += "Details: \n"
                for val in unpacked:
                    errorMessage += "\t{}\n".format(val)
            raise Exception(errorMessage) from None
    except ValueError:
        # maybe report back to wren in the future
        pass
    errorMessage += "\tError Message: {}\n".format(rpc_error.details())
    raise Exception(errorMessage) from None


def unpack_details(grpc_detail):
    """Unpack a grpc status detail field (which is a 'google.protobuf.Any' type).
    `Unpack()` checks the descriptor of the passed-in message object against the stored one
    and returns False if they don't match and does not attempt any unpacking; True otherwise.
        Source:
            https://github.com/protocolbuffers/protobuf/blob/master/python/google/protobuf/internal/well_known_types.py#L81
            https://github.com/protocolbuffers/protobuf/blob/master/python/google/protobuf/internal/python_message.py#L1135

        Raises:
            google.protobuf.message.DecodeError: If it can't deserialize the message object

    `Is()` checks if the `Any` message represents the given protocol buffer type.
    """
    if grpc_detail.Is(error_details_pb2.BadRequest.DESCRIPTOR):
        val = error_details_pb2.BadRequest()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.PreconditionFailure.DESCRIPTOR):
        val = error_details_pb2.PreconditionFailure()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.RetryInfo.DESCRIPTOR):
        val = error_details_pb2.RetryInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.DebugInfo.DESCRIPTOR):
        val = error_details_pb2.DebugInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.QuotaFailure.DESCRIPTOR):
        val = error_details_pb2.QuotaFailure()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.RequestInfo.DESCRIPTOR):
        val = error_details_pb2.RequestInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.ResourceInfo.DESCRIPTOR):
        val = error_details_pb2.ResourceInfo()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.Help.DESCRIPTOR):
        val = error_details_pb2.Help()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(error_details_pb2.LocalizedMessage.DESCRIPTOR):
        val = error_details_pb2.LocalizedMessage()
        grpc_detail.Unpack(val)
        return val
    elif grpc_detail.Is(kaskada_error_details_pb2.FenlDiagnostics.DESCRIPTOR):
        val = kaskada_error_details_pb2.FenlDiagnostics()
        grpc_detail.Unpack(val)
        error_msgs = []
        for fenl_diagnostic in val.fenl_diagnostics:
            error_msgs.append(fenl_diagnostic.formatted.replace("\n", "\n\t"))
        return "\n".join(error_msgs)
    else:
        raise ValueError(grpc_detail.type_url)


def handleException(e: Exception):
    raise Exception("An exception occurred: {}".format(e)) from None


def validate_client(client: Client):
    if client is None:
        raise ValueError("Client must be provided")
    if client.computeStub is None:
        raise ValueError(
            "Invalid client provided. Compute service stub was not initialized properly."
        )
    if client.viewStub is None:
        raise ValueError(
            "Invalid client provided. View service stub was not initialized properly."
        )
    if client.tableStub is None:
        raise ValueError(
            "Invalid client provided. Table service stubs was not initialized properly."
        )



"""
Copyright (C) 2021 Kaskada Inc. All rights reserved.

This package cannot be used, copied or distributed without the express
written permission of Kaskada Inc.

For licensing inquiries, please contact us at info@kaskada.com.
"""

from datetime import datetime, timedelta
from typing import Tuple, List

import grpc
import http.client
import json
import certifi

import kaskada.api.v1alpha.compute_pb2_grpc as compute_grpc
import kaskada.api.v1alpha.materialization_pb2_grpc as material_grpc
import kaskada.api.v1alpha.table_pb2_grpc as table_grpc
import kaskada.api.v1alpha.staged_file_pb2_grpc as staging_grpc
import kaskada.api.v1alpha.view_pb2_grpc as view_grpc


class Client(object):
    LEGACY_ENGINE = "LEGACY"

    def getBearerToken(
        endpoint: str, audience: str, client_id: str, client_secret: str
    ):
        """
        Requests a bearer token given the provided parameters

        Args:
            endpoint (str): the authentication endpoint to request a token from
            audience (str): the audience the token is requesting
            client_id (str): the user provided client id
            client_secret (str): the user provided client secret

        Returns:
            AccessToken: Kaskada Access Token
        """
        conn = http.client.HTTPSConnection(endpoint)
        payload = {
            "client_id": client_id,
            "client_secret": client_secret,
            "audience": audience,
            "grant_type": "client_credentials",
        }
        headers = {
            "content-type": "application/json",
        }
        conn.request("POST", "/oauth/token", json.dumps(payload), headers)
        res = conn.getresponse()
        data = res.read().decode("utf-8")
        return Client.AccessToken.fromJson(data, endpoint, audience, client_id, client_secret)

    class AccessToken(object):
        def fromJson(
            json_data: str,
            endpoint: str,
            audience: str,
            client_id: str,
            client_secret: str,
        ):
            """
            AccessToken constructor

            Args:
                json_data (str): UTF-8 response from the oauth token request
                endpoint (str): the authentication endpoint to request a token from
                audience (str): the audience the token is requesting
                client_id (str): the user provided client id
                client_secret (str): the user provided client secret

            Raises:
                PermissionError: Unable to validate access token.
            """
            data = json.loads(json_data)
            if (
                "access_token" not in data
                and "expires_in" not in data
                and "token_type" not in data
            ):
                raise PermissionError(
                    "Unable to validate access token. Token details: {}".format(
                        json_data
                    )
                )
            access_token = data["access_token"]
            # The token is valid from now + expires_in seconds.
            # Set the renewal time to one hour before the expires time for safety.
            expires_at = datetime.now() + timedelta(0, data["expires_in"] - 3600)
            token_type = data["token_type"]
            return Client.AccessToken(
                access_token=access_token,
                expires_at=expires_at,
                token_type=token_type,
                endpoint=endpoint,
                audience=audience,
                client_id=client_id,
                client_secret=client_secret)

        def fromJwt(
            jwt: str,
        ):
            """
            Creates an AccessToken from a JWT

            Args:
                jwt (str): A pre-authorized JWT

            """
            return Client.AccessToken(access_token=jwt)

        def __init__(self,
                     access_token: str = None,
                     expires_at: int = None,
                     token_type: str = None,
                     endpoint: str = None,
                     audience: str = None,
                     client_id: str = None,
                     client_secret: str = None):
            self.init_time = datetime.now()
            self.access_token = access_token
            self.expires_at = expires_at
            self.token_type = token_type
            self.endpoint = endpoint
            self.client_id = client_id
            self.client_secret = client_secret
            self.audience = audience

    def authorized(
        client_id: str,
        client_secret: str,
        exchange_endpoint: str,
        audience: str,
        endpoint: str,
        is_secure: bool,
    ):
        """
        Initializes an authorized Kaskada Client

        Args:
            client_id (str): Kaskada Client ID
            client_secret (str): Kaskada Client Secret
            exchange_endpoint (str): Authentication endpoint to exchange JWTs
            audience (str): Authentication audience
            endpoint (str): API Endpoint
            is_secure (bool): Use SSL connection.

        Returns:
            Client: Authorized client
        """
        token = Client.getBearerToken(
            exchange_endpoint, audience, client_id, client_secret
        )
        return Client(client_id, token, endpoint, is_secure)

    def preauthorized(
        jwt: str,
        endpoint: str,
        is_secure: bool
    ):
        """
        Initializes a preauthorized Kaskada Client

        Args:
            jwt (str): A valid JWT
            endpoint (str): API Endpoint
            is_secure (bool): Use SSL connection.

        Returns:
            Client: Preauthorized client
        """
        token = Client.AccessToken.fromJwt(jwt)
        return Client(None, token, endpoint, is_secure)

    def client_id_only(client_id: str, endpoint: str, is_secure: bool):
        """
        Initializes a Kaskada Client with a Client ID only

        Args:
            client_id (str): Kaskada only Client ID
            endpoint (str):  API Endpoint
            is_secure (bool): Use SSL

        Returns:
            Client: Demo client
        """
        return Client(client_id, None, endpoint, is_secure)

    def __init__(self,
                 client_id: str = None,
                 client_secret: str = None,
                 jwt: str = None,
                 exchange_endpoint: str = None,
                 audience: str = None,
                 endpoint: str = None,
                 is_secure: bool = None):

        token: Client.AccessToken = None
        # Client with a pre-authorized JWT (using cached JWTs)
        if client_id is None and client_secret is None:
            assert jwt is not None, "client initialization requires a client ID + client secret or a jwt"
            token = Client.AccessToken.fromJwt(jwt)
        # Client with credentials needs to exchange for a token
        elif client_id is not None and client_secret is not None:
            token = Client.getBearerToken(
                exchange_endpoint, audience, client_id, client_secret
            )
        # Client ID only for demo and trial users
        else:
            assert client_id is not None, "client ID must be provided"

        assert exchange_endpoint is not None and len(exchange_endpoint) > 0, \
            "invalid authorization exchange endpoint provided"
        assert audience is not None and len(audience) > 0, "invalid authorization audience provided"
        assert endpoint is not None and len(endpoint) > 0, "invalid api endpoint provided"
        assert is_secure is not None, "invalid is_secure flag provided"

        if is_secure:
            with open(certifi.where(), "rb") as f:
                trusted_certs = f.read()
            credentials = grpc.ssl_channel_credentials(root_certificates=trusted_certs)
            channel = grpc.secure_channel(endpoint, credentials)
        else:
            channel = grpc.insecure_channel(endpoint)
        self.computeStub = compute_grpc.ComputeServiceStub(channel)
        self.tableStub = table_grpc.TableServiceStub(channel)
        self.viewStub = view_grpc.ViewServiceStub(channel)
        self.stagingStub = staging_grpc.StagedFileServiceStub(channel)
        self.materializationStub = material_grpc.MaterializationServiceStub(channel)
        self.client_id = client_id
        self.token = token

    def get_metadata(self) -> List[Tuple[str, str]]:
        """
        Fetches the metadata for the current client. Renews token if necessary.

        Raises:
            Exception: invalid token

        Returns:
            List[Tuple[str, str]]: Client metadata
        """
        if self.token is None:
            return [("client-id", self.client_id)]
        elif self.token.client_id is not None and self.token.client_secret is not None:
            if self.token.expires_at is None:
                raise Exception(
                    "invalid token. please re-initialize the Kaskada client for continued use"
                )

            current_time = datetime.now()
            if self.token.expires_at is not None:
                remaining_token_time = (
                    self.token.expires_at - current_time
                ).total_seconds()
                # A token is expired if the difference between the expired time and current time is non-positive
                if remaining_token_time <= 0:

                    # Refresh the token by requesting a new one
                    self.token = Client.getBearerToken(
                        self.token.endpoint,
                        self.token.audience,
                        self.token.client_id,
                        self.token.client_secret,
                    )
        return [("authorization", self.token.access_token)]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kaskada",
    version="0.0.20",
    author="Kaskada",
    author_email="support@kaskada.com",
    description="A client library for the Kaskada time travel machine learning service",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://kaskada.com",
    project_urls={
        "Documentation": "https://docs.kaskada.com/",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Jupyter",
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
    ],
    package_dir={"": "src"},
    packages=["kaskada"],
    package_data={"kaskada": ["formatters.css", "formatters.js"]},
    python_requires=">=3.6",
    install_requires=[
        'certifi',
        'domonic',
        'googleapis-common-protos',
        'grpcio~=1.47.0',
        'grpcio-status',
        'html5lib',
        'ipython',
        'kaskada_grpc==0.0.14',
        'pandas',
        'protobuf',
        'pyarrow',
        'requests',
    ],
)
